/* Copyright (C) 2023 Artyom V. Poptsov <poptsov.artyom@gmail.com>
 *
 * This file is part of Guile-SSH.
 *
 * Guile-SSH is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Guile-SSH is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Guile-SSH.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __CALLBACKS_H__
#define __CALLBACKS_H__

int callback_set_p (SCM callbacks, const char* name);
SCM callback_ref (SCM callbacks, const char* name);
void callback_validate (SCM parent, SCM callbacks, const char* name);
SCM callback_userdata_ref (SCM callbacks);

#endif /* __CALLBACKS_H__ */

/* callbacks.h ends here. */
